--[[ <HCExtension>
@name			Content Master
@author			nadenj@mail.ru
@version		0.10
@description	Content Editing Engine
@event			Init/init
@event			BeforeAnswerHeaderSend/header
@event			Options/options
@event			Destroy/destroy
</HCExtension> ]]

--  Shift -    
-- *********************************************************************************************

local function to_log(...)
	if hc_static.CMLog then
		hc.put_to_log(...)
	end
end

function init()
	hc_static.CMData = {}
	hc_static.CMLists = {}
	local ini_fname = re.replace(hc.script_name, [[(.*\.).*]], [[\1ini]])
	dofile(ini_fname)
	local divide_rank = 200 --    200 

	local function DivideLST(List, rank)
		local num = 1
		local ind1 = 1
		local ind2 = #List
		local parts = {}
		while ind1+rank-1 < ind2 do
			parts[num] = '(?:' .. table.concat(List, '|', ind1, ind1+rank-1) .. ')'
			num = num + 1
			ind1 = ind1 + rank
		end
		parts[num] = '(?:' .. table.concat(List, '|', ind1) .. ')'
		return parts
	end

	local function toregexp(str)
		if str:sub(1,1) == '+' then
			str = str:sub(2)
			str = re.replace(str, [=[[?.+|)(}{\][\\$^]]=], [[\\\0]], true)
			str = re.replace(str, [[^\*+|\*+$]], '', true)	--     '*'
			str = re.replace(str, [[\*+]], '.*', true)
		end
		return str
	end

	for i,v in ipairs(hc_static.CMData) do
-- ---------------------------   Adblock -----------------------
		if (v.Type=='Adblock') and v.File then
			v.Name = v.Name or ' AdblockPlus'
			local AdblockPlus_Black_List = {}
			local AdblockPlus_White_List = {}
			local AdblockPlus_CSS_List = ''
			if type(v.File) == 'string' then
				v.File = {v.File}
			end
			for _,file in ipairs(v.File) do	--     
				for line in io.lines(re.replace(hc.script_name, [[\\[^\\]+$]], [[\]]) .. file) do
					local regexp = re.find(line, [[^(?|(?:@@)?/(.*)/|(?![![])(?:@@)?([^$#]*+))(?:$|\$(.+))]], 1)
					if regexp then		--     URL
						local opts = re.find(nil, nil, 2) or ''
						--    " "  
						if not re.match(line, [[^(?:@@)?/.*/($|\$)]]) then
							regexp = toregexp('+' .. regexp)
							regexp = re.replace(regexp, [[\^]], [[(?:[^\w.%-]|_|$)]], true)
							regexp = re.replace(regexp, [[^\\\|\\\|]], [[^https?://(www\.)?]])
							regexp = re.replace(regexp, [[^\\\|]], '^')
							regexp = re.replace(regexp, [[\\\|$]], '$')
						end
						for option in opts:gmatch('[^,]+') do		--  
							local domains = re.find(option, [[^domain=\K.+]])
								if domains then
									local lst = ''
									local neg_lst = ''
									for current_dom in domains:gmatch('[^\|]+') do
										current_dom = current_dom:gsub('%.', [[\.]])
										if current_dom:find('^~') then
											neg_lst = neg_lst .. re.find(current_dom, [[^~\K.+]]) .. '|'
										else
											lst = lst .. current_dom .. '|'
										end
									end
									if neg_lst~='' then
										regexp = [[^(?!\w+:(?://)?[^?:/]*?(?:]] .. re.find(neg_lst, [[.+(?=\|$)]]) .. ')).*?(?:' .. regexp .. ')'
									end
									if lst~='' then
										regexp = [[^(?=\w+:(?://)?[^?:/]*?(?:]] .. re.find(lst, [[.+(?=\|$)]]) .. ')).*?(?:' .. regexp .. ')'
									end
								elseif option == 'match-case' then
									regexp = '(?-i:' .. regexp .. ')'
								elseif option == 'stylesheet' then
									regexp = [[^(?=[^?]*?\.css($|\?)).*?(?:]] .. regexp .. ')'
								elseif option == '~stylesheet' then
									regexp = [[^(?![^?]*?\.css($|\?)).*?(?:]] .. regexp .. ')'
								else	--      
									regexp = nil
									break
								end
						end
						if regexp and regexp~='' then
								if line:find('^@@') then
									table.insert(AdblockPlus_White_List, regexp)
								else
									table.insert(AdblockPlus_Black_List, regexp)
								end
						end
					else	--     CSS
						regexp = re.find(line, [[^\#\#\K.*]], 0)
						if regexp then
							AdblockPlus_CSS_List = AdblockPlus_CSS_List .. '   ' .. regexp .. ' {display: none !important;}\r\n'
						else
							regexp = re.find(line, [[^\#\K[^#].*]], 0)
							if regexp then
								regexp = re.replace(regexp, [[([^(]++)\(([^=)]++)\)]], [[\1#\2, \1.\2]], true)
								regexp = re.replace(regexp, [[\(([^)]++)\)]], [=[[\1]]=], true)
								AdblockPlus_CSS_List = AdblockPlus_CSS_List .. '   ' .. regexp .. ' {display: none !important;}\r\n'
							end
						end
					end -- if regexp
				end -- for line
			end	-- for _,file
			v.Rules = v.Rules or {}
			if AdblockPlus_Black_List[1] then
				table.insert(v.Rules, {
					Name = ' HTML-  URL  AdblockPlus',
					Type = 'Adblock',
					Find = [[<!--(?:(?:.)*?-->(*SKIP)|(*COMMIT))(*F)|(?=<(?:[^>])+?\s(?:href|src|data|value)=['"]?+(?<hc_subfind>(?:[^'">\s])*+).)(?&html_element)]],
					Replace = '<!-- ********** CONTENT MASTER: Element deleted by Adblock ********* -->',
					__WhiteSubFinds = DivideLST(AdblockPlus_White_List, divide_rank),
					__SubFinds = DivideLST(AdblockPlus_Black_List, divide_rank),
--					SubFind = '^(?!.*?(?:' .. table.concat(AdblockPlus_White_List, '|') .. ')).*?(?:' .. table.concat(AdblockPlus_Black_List, '|') .. ')',
					Counted = true,
					On = true,
				})
			end
			if AdblockPlus_CSS_List ~= '' then
				table.insert(v.Rules, {
					Name = ' HTML-   CSS  AdblockPlus',
					Find=[[(?=</head>)]],
					Replace = '\r\n\r\n<!-- INSERTED BY CONTENT MASTER BEGIN: -->\r\n' ..
								 '<style type="text/css">\r\n' .. AdblockPlus_CSS_List .. '</style>\r\n' ..
								 '<!-- INSERTED BY CONTENT MASTER END -->\r\n\r\n',
					Counted = false,
					Once = true,
					On = true,
				})
			end
-- ---------------------------       -----------------------
		elseif v.Type=='HandyCache' and not v.Find then
			v.Name = v.Name or ' HandyCache'
			local HandyCache_Black_List = {}		--  
			local HandyCache_White_List = {}		--  
			local regexp, regexp2
			for line in io.lines(hc.ini_path .. 'black.lst') do
				local t = re.match(line, [[^True\#~\#(?!ip\s*+\d)(.*?)\#~\#(?!ip\s*+\d)(.*?)\#~\#\#~\#\d*$]])
				if t then
					regexp  = toregexp(line:sub(t[1].begin_pos, t[1].end_pos))
					regexp2 = toregexp(line:sub(t[2].begin_pos, t[2].end_pos))
					regexp = (regexp2~='') and ('^(?!.*?(?:' .. regexp2 .. ')).*?(?:' .. regexp .. ')') or regexp
					if regexp and regexp~='' then
						table.insert(HandyCache_Black_List, regexp)
					end
				end
			end
			for line in io.lines(hc.ini_path .. 'white.lst') do
				local t = re.match(line, [[^(?>True\#~\#(?!ip\s*+\d)(.*?)\#~\#)(?>(?!ip\s*+\d)(.*?)\#~\#)\#~\#True\#~\#]])
				if t then
					regexp  = toregexp(line:sub(t[1].begin_pos, t[1].end_pos))
					regexp2 = toregexp(line:sub(t[2].begin_pos, t[2].end_pos))
					regexp = (regexp2~='') and ('^(?!.*?(?:' .. regexp2 .. ')).*?(?:' .. regexp .. ')') or regexp
					if regexp and regexp~='' then
						table.insert(HandyCache_White_List, regexp)
					end
				end
			end
			v.Rules = v.Rules or {}
			table.insert(v.Rules, {
				Name = ' HTML-  URL    ',
				Type = 'HandyCache',
				Find = [[<!--(?:(?:.)*?-->(*SKIP)|(*COMMIT))(*F)|(?=(?><(?:[^>])+?\s(?:href|src|data|value)=['"]?+(?=https?:|ftp:|(?!(?:\w)++:)))(?<hc_subfind>(?:(?:[^'">\s]){5})*+(?:[^'">\s])*+).)(?&html_element)|<script\b(?:(?:.)*?</script>(*SKIP)|(*COMMIT))(*F)]],
				Replace = '<!-- ********** CONTENT MASTER: Element deleted by HandyCache Black List. URL=\\1 -->',
				__WhiteSubFinds = DivideLST(HandyCache_White_List, divide_rank),
				__SubFinds = DivideLST(HandyCache_Black_List, divide_rank),
--				SubFind = '^(?!.*?(?:' .. table.concat(HandyCache_White_List, '|') .. ')).*?(?:' .. table.concat(HandyCache_Black_List, '|') .. ')',
				Counted = true,
				On = true,
			})
			table.insert(v.Rules, {
				Name = '   URL    ',
				Type = 'HandyCache',
				Find = [[<!--(?:(?:.)*?-->(*SKIP)|(*COMMIT))(*F)|<script\b(?:[^>])*?>(?:(?!</script>).)*?(?<hc_subfind>http://(?:[^<'">\s])*+)(?:.)*?</script>]],
				Replace = '<!-- ********** CONTENT MASTER: Script deleted by HandyCache Black List. URL=\\1 -->',
				__WhiteSubFinds = DivideLST(HandyCache_White_List, divide_rank),
				__SubFinds = DivideLST(HandyCache_Black_List, divide_rank),
--				SubFind = '^(?!.*?(?:' .. table.concat(HandyCache_White_List, '|') .. ')).*?(?:' .. table.concat(HandyCache_Black_List, '|') .. ')',
				Counted = true,
				On = true,
			})
		end	-- if v.Type=='Adblock'
	end	-- for i,v

-- -------------------------     --------------------------------
	for i,v in ipairs(hc_static.CMData) do
		for _,rule in ipairs(v.Rules) do
			for j,w in pairs(hc_static.CMLists) do
				local lst = re.replace(w, [[\A[\r\n]++|[\r\n]++\z]], '', true)  --  .  .  
				lst = re.replace(lst, [[[\r\n]++]], '|', true)
				lst = re.replace(lst, [[\|\|++]], '|', true)
				rule.Find = rule.Find and re.replace(rule.Find, [[\(?&]] .. j .. [[\).*]], [[\0(?(DEFINE)(?<]] .. j .. '>' .. lst .. '))', false)
				rule.SubFind = rule.SubFind and re.replace(rule.SubFind, [[\(?&]] .. j .. [[\).*]], [[\0(?(DEFINE)(?<]] .. j .. '>' .. lst .. '))', false)
			end
		end
	end
-- -------------------------     Content Master.opt --------------------------------
	local f = io.open(opt_file_name(), 'rb')
	local s, sfile
	if f then 
		sfile = f:read("*a")
		f:close()
	end
	sfile = sfile or ''
	local H = {}
	for i,v in ipairs(hc_static.CMData) do
		H[i] = {}
		s= re.find(sfile, toregexp(v.Name) .. '=(true|false)', 1)
		v.On = tostring(s or v.On)=='true'
		for j,rule in ipairs(v.Rules) do
			if rule.Counted == nil then		--    ,   
				rule.Counted = v.Counted
			end
			s= re.find(sfile, toregexp(rule.Name) .. '=(true|false)(?:,(\\d+))?', 1)
			if not s and rule.On==nil then
				rule.On = v.On
			elseif s then
				rule.On = s=='true'
			end
			s = re.find(nil, nil, 2)
			H[i][j] = tonumber(s or 0)
		end
	end
	hc.set_global('CM_Hits', H)
	s= re.find(sfile, '   =(true|false)', 1)
    hc_static.CMLog = s=='true'
	to_log('Content Master: initialized from ', ini_fname)
end


-- *********************************************************************************************

function header()
	-- to_log('\r\nANSWER HEADER HANDLER STARTED...')
	if re.match(hc.answer_header, [[\A\S++\s++200\s.*?^Content-Type:\s*+text/html]]) and	--   200   html
		not re.match(hc.get_keyboard_state(), 'VK_RSHIFT') then										--     Shift

		hc.call_me_for('BeforeAnswerBodySend', 'body')		--    
		-- to_log('   ')
	end
	-- to_log('\r\nANSWER HEADER HANDLER FINISHED')
end

-- *********************************************************************************************

function body()

	to_log('\r\n  ...')
	local ToNextItem = hc.answer_body
	CM = CM or {Freezed={}, ReplCount=0, ChunkNumber=0, In=0, Out=0}
	CM.ChunkNumber = CM.ChunkNumber + 1	--    
	CM.In = CM.In + #ToNextItem
-- to_log('\r\n 80   :\r\n', ToNextItem:sub(1,80))
-- to_log('\r\n 150   :\r\n', ToNextItem:sub(-150))
	local replaces_in_chunk = 0	--     
	local trapped = 0			--    
	local Subj, t, PCRE_ErrMsg

	local url = hc.url
	local host = re.find(url, [[^\w+://[^:/?]*]])
	local domain = re.find(host, [[^\w+://\K.*]])

	local function get_capture(s)
		local n = tonumber(s:sub(2,2))
		return Subj:sub(t[n].begin_pos, t[n].end_pos)
	end
	
	for i,v in ipairs(hc_static.CMData) do
		if not hc.last_part and ToNextItem=='' then break end
		if v.On then		--    ,   
			CM.Freezed[i] = CM.Freezed[i] or {}
			for j,rule in ipairs(v.Rules) do
				if not hc.last_part and ToNextItem=='' then break end
				if rule.On and not rule.__Off then	--   ()  ,   
					Subj = (CM.Freezed[i][j] or '') .. ToNextItem
					local start_pos = 1
					ToNextItem = Subj:sub(1, start_pos-1)
					local log_msg = ''
					t, PCRE_ErrMsg = re.match(Subj, rule.Find, start_pos, true)
					local k = (rule.SubFind or rule.__SubFinds) and re.named_substr_index('hc_subfind')
					while t do
						if rule.SubFind or rule.__SubFinds then
							local Subj2 = Subj:sub(t[k].begin_pos, t[k].end_pos)
							if v.Type=='Adblock' or v.Type=='HandyCache' then
								--  url ,    
								if Subj2:sub(1,1)=='/' then
									Subj2 = host .. Subj2
								elseif not Subj2:find('^%w+:') then
									Subj2 = host .. '/' .. Subj2
								end
							end
		-- hc.put_to_log(Subj:sub(t[0].begin_pos, t[0].end_pos), '\r\n', v.Name, '\r\n', Subj2)
							--     URL 
							if v.Type=='HandyCache' then
								Subj2 = re.replace(Subj2, [[^(http://(www\.)?)?]], '')
							end
							local t2, PCRE_ErrMsg2
							if v.Type=='Adblock' or v.Type=='HandyCache' then
								local operate = true
								for _,vv in ipairs(rule.__WhiteSubFinds) do
									t2, PCRE_ErrMsg2 = re.match(Subj2, vv)
									if t2 then
										operate = false
										t2 = nil
										PCRE_ErrMsg2 = 'PCRE_ERROR_NOMATCH'
										break
									end
								end
								if operate then
									for _, vv in ipairs(rule.__SubFinds) do
										t2, PCRE_ErrMsg2 = re.match(Subj2, vv)
										if t2 then break end
									end
								end
							else
								t2, PCRE_ErrMsg2 = re.match(Subj2, rule.SubFind)
							end
							if t2 then
								local repl = rule.Replace:gsub('\\%d', get_capture)
								ToNextItem = ToNextItem .. Subj:sub(start_pos, t[0].begin_pos-1) .. repl
								local H = hc.get_global('CM_Hits')
								H[i][j] = H[i][j] + 1
								hc.set_global('CM_Hits', H)
								replaces_in_chunk = replaces_in_chunk + (rule.Counted and 1 or 0)
								log_msg = log_msg .. '\r\n\r\n:\r\n' .. Subj:sub(t[0].begin_pos, t[0].end_pos) ..
									'\r\n :\r\n' .. repl
								start_pos = t[0].end_pos + 1
								if rule.Once then
									rule.__Off = true
									t = nil
									break
								end
							elseif PCRE_ErrMsg2:find('PCRE_ERROR_NOMATCH', 1, true) then
								ToNextItem = ToNextItem .. Subj:sub(start_pos, t[0].begin_pos)
								start_pos = t[0].begin_pos + 1
							else
								log_msg = log_msg .. '\r\n PCRE   SubFind: ' .. PCRE_ErrMsg2 .. '\r\nSubFind = \r\n' .. (rule.SubFind or rule.__SubFinds)
								hc.put_msg('Content Master: ', rule.Name or rule.Find, '\r\nSubFind: ', PCRE_ErrMsg2)
								break
							end
							t, PCRE_ErrMsg = re.match(Subj, rule.Find, start_pos, true)
						else
							local repl = rule.Replace:gsub('\\%d', get_capture)
							ToNextItem = ToNextItem .. Subj:sub(start_pos, t[0].begin_pos-1) .. repl
							local H = hc.get_global('CM_Hits')
							H[i][j] = H[i][j] + 1
							hc.set_global('CM_Hits', H)
							replaces_in_chunk = replaces_in_chunk + (rule.Counted and 1 or 0)
							log_msg = log_msg .. '\r\n\r\n:\r\n' .. Subj:sub(t[0].begin_pos, t[0].end_pos) ..
								'\r\n :\r\n' .. repl
							start_pos = t[0].end_pos + 1
							if rule.Once then	--       ,    
								rule.__Off = true
								t = nil
								break
							end
							t, PCRE_ErrMsg = re.match(nil, nil, start_pos, true)
						end
					end	-- while t do
					if PCRE_ErrMsg and PCRE_ErrMsg:find('PCRE_ERROR_PARTIAL', 1, true) then
						CM.Freezed[i][j] = Subj:sub(start_pos)
					else
						ToNextItem = ToNextItem .. Subj:sub(start_pos)
						CM.Freezed[i][j] = ''
						if PCRE_ErrMsg and not PCRE_ErrMsg:find('PCRE_ERROR_NOMATCH', 1, true) then
							log_msg = log_msg .. '\r\n   Find: ' .. PCRE_ErrMsg
							hc.put_msg('Content Master: ', v.Name or '', '\r\n', rule.Name or rule.Find, ': ', PCRE_ErrMsg)
						end
					end
					if hc.last_part then
						ToNextItem = ToNextItem .. CM.Freezed[i][j]
						CM.Freezed[i][j] = ''
						rule.__Off = nil
					end
					trapped = trapped + #CM.Freezed[i][j]
					if log_msg ~= '' then
						to_log(': ', v.Name or '', ' / ', rule.Name or rule.Find, log_msg)
					end
				end	--   ()  
			end -- for j,rule
		end	--    
	end	-- for i,v
	
	CM.ReplCount = CM.ReplCount + replaces_in_chunk
	CM.Out = CM.Out + #ToNextItem
	to_log('\r\n=============   ', CM.ChunkNumber, hc.last_part and ' ()' or '',
					' /  ============',
					'\r\n,      : ', #hc.answer_body, ' / ', CM.In,
					'\r\n     : ', replaces_in_chunk, ' / ', CM.ReplCount,
					'\r\n,      : ', #ToNextItem, ' / ', CM.Out,
					'\r\n,  : ', trapped,
					'\r\n============================================')
	if hc.last_part then
		hc.monitor_string = 'CM: ' .. CM.ReplCount .. ' hits'
		CM = nil
	end
	hc.answer_body = ToNextItem
	to_log('\r\n   ')
end


function opt_file_name()
  return re.replace(hc.script_name, [[(.*\.).*]], [[\1opt]])
end

function SaveParams()  
	local f = assert(io.open(opt_file_name(), 'wb'))
	if not f then return end
	local H = hc.get_global('CM_Hits')
	for i,v in ipairs(hc_static.CMData) do
		f:write(v.Name .. '=' .. tostring(v.On) .. '\r\n')
		for j,rule in ipairs(v.Rules) do
			f:write(rule.Name .. '=' .. tostring(rule.On) .. ',' .. (H[i][j] or 0) .. '\r\n')
		end
	end
	f:write('   =' .. tostring(hc_static.CMLog))
	f:close()
end  


function options()
	require "vcl"
	if Form then
		Form:Free()
		Form=nil
	end
	Form = VCL.Form('Form')
	local x,y,w,h= hc.window_pos()
	Form._ = { Caption=' Content Master', width=580 }
	cb = {}
	cbrule = {}
	local top_pos = 20
	local H = hc.get_global('CM_Hits')
	for i,v in ipairs(hc_static.CMData) do
		cb[i] = VCL.CheckBox(Form, 'cb'..i)
		cb[i]._= { caption=v.Name, top=top_pos, left=30, width=500, Checked=v.On}
		top_pos = top_pos + 20
		cbrule[i] = {}
		for j,rule in ipairs(v.Rules) do
			cbrule[i][j] = VCL.CheckBox(Form, 'cbrule'..i.. '_' ..j)
			cbrule[i][j]._= { caption=rule.Name .. (rule.Counted and (' (' .. H[i][j] .. ')') or ''), top=top_pos, left=60, width=500, Checked=rule.On}
			top_pos = top_pos + 20
		end
	end
	cbLog = VCL.CheckBox(Form, "cbLog")
	cbLog._= { caption='   ', top=top_pos+20, left=30, width=200, Checked=hc_static.CMLog}

	Form._ = { height=top_pos+130 }
	Form._ = { left=x+(w-Form.width)/2, top=y+(h-Form.height)/2, OnResize='OnFormResize' }

	OkButton = VCL.Button(Form, "OkButton")
	OkButton._ = {onclick = "onOkButtonClick", width=100, left=30, caption = "OK", top= Form.clientheight-OkButton.height-10}
	CancelButton = VCL.Button(Form, "CancelButton")
	CancelButton._ = {onclick = "onCancelButtonClick", width=100, left=140, top=OkButton.top, caption = "Cancel"}
	Form:ShowModal()
	Form:Free()
	Form = nil
end

function onOkButtonClick(Sender)
	for i,v in ipairs(hc_static.CMData) do
		v.On = cb[i].Checked
		for j,rule in ipairs(v.Rules) do
			rule.On = cbrule[i][j].Checked
		end
	end
	hc_static.CMLog = cbLog.Checked
  Form:Close()
  SaveParams()
end

function onCancelButtonClick(Sender)
  Form:Close()
end


function destroy()
	SaveParams()
	hc.set_global('CM_Hits')
end
